// *************************************************************************************************
//
//      Copyright (C) 2009 Texas Instruments Incorporated - http://www.ti.com/
//
//
//        Redistribution and use in source and binary forms, with or without
//        modification, are permitted provided that the following conditions
//        are met:
//
//          Redistributions of source code must retain the above copyright
//          notice, this list of conditions and the following disclaimer.
//
//          Redistributions in binary form must reproduce the above copyright
//          notice, this list of conditions and the following disclaimer in the
//          documentation and/or other materials provided with the
//          distribution.
//
//          Neither the name of Texas Instruments Incorporated nor the names of
//          its contributors may be used to endorse or promote products derived
//          from this software without specific prior written permission.
//
//        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//        "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//        LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//        A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//        OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//        SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//        LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//        DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//        THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//        OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// *************************************************************************************************
// Temperature measurement functions.
// *************************************************************************************************

// *************************************************************************************************
// Include section

// system
#include "project.h"

// driver
#include "display.h"
#include "bmp_as.h"
#include "cma_as.h"
#include "as.h"

// logic
#include "acceleration.h"
#include "simpliciti.h"
#include "user.h"

#include <string.h>

#define BIKE_AXIS ACCEL_X

// *************************************************************************************************
// Global Variable section
struct accel sAccel;

// The default for level ground, assuming the inclinometer is not mounted on the bike at an angle
s16 level_ground[3] = {0, 0, 0};

// To smooth out fluctuations in the acceleration (including the effects of pedalling)
// sum the samples over a few seconds. we'll display the average. A sample is read approximately
// every 16 ms.
#define MOVING_SUM_SAMPLE_COUNT 200
s16 moving_sum_samples[MOVING_SUM_SAMPLE_COUNT];
u16 moving_sum_index = 0u;
s32 moving_sum;

s32 max_climb, max_descent;

// Conversion values from data to mgrav taken from BMA250 datasheet (rev 1.05, figure 4)
	const s16 bmp_mgrav_per_bit[7] = { 16, 31, 63, 125, 250, 500, 1000 };
// Conversion values from data to mgrav taken from CMA3000-D0x datasheet (rev 0.4, table 4)
	const s16 cma_mgrav_per_bit[7] = { 18, 36, 71, 143, 286, 571, 1142 };

// *************************************************************************************************
// Extern section

extern s16 convert_acceleration_value_to_tenthsOfDegrees(s8 value);

// *************************************************************************************************
// @fn          reset_acceleration
// @brief       Reset acceleration variables.
// @param       none
// @return      none
// *************************************************************************************************
void reset_acceleration(void)
{
    // Use the axis that is parallel to the forward motion of the bike
    sAccel.view_style = BIKE_AXIS;

    // Clear timeout counter
    sAccel.timeout = 0u;

    // Default mode is off
    sAccel.mode = ACCEL_MODE_OFF;

    calibrate_acceleration(1u);

    memset(moving_sum_samples, 0, sizeof(moving_sum_samples));
    moving_sum = 0;
    moving_sum_index = 0u;
    max_climb = max_descent = 0;

}

// *************************************************************************************************
// @fn          calibrate_acceleration
// @brief       Acceleration direct function. Long press of the Star button calibrates the level
// @param       u8 line         LINE2
// @return      none
// *************************************************************************************************
void calibrate_acceleration(u8 line)
{
    // Get data from sensor
	if (bmp_used)
	{
        bmp_as_get_data((s8 *) sAccel.xyz);
	}
	else
	{
        cma_as_get_data((s8 *) sAccel.xyz);
	}

	level_ground[0] = convert_acceleration_value_to_tenthsOfDegrees(sAccel.xyz[0]);
	level_ground[1] = convert_acceleration_value_to_tenthsOfDegrees(sAccel.xyz[1]);
	level_ground[2] = convert_acceleration_value_to_tenthsOfDegrees(sAccel.xyz[2]);
	max_climb = max_descent = 0;
	memset(moving_sum_samples, 0, sizeof(moving_sum_samples));
	moving_sum_index = 0;
	moving_sum = 0;
}

// *************************************************************************************************
// @fn          sx_acceleration
// @brief       Acceleration direct function. Button UP switches between current/max climb/max descent values.
// @param       u8 line         LINE2
// @return      none
// *************************************************************************************************
void sx_acceleration(u8 line)
{
    if (++sAccel.view_style > 2)
        sAccel.view_style = 0;
}

// *************************************************************************************************
// @fn          convert_acceleration_value_to_mgrav
// @brief       Converts measured value to mgrav units
// @param       s8 value        g data from sensor
// @return      s16                     Acceleration (mgrav)
// *************************************************************************************************
s16 convert_acceleration_value_to_mgrav(s8 value)
{
    s16 result;
    u8 i;
    s8 sign = 1;

    if (value < 0)
    {
        value = -value;
        sign = -1;
    }

    result = 0;
    for (i = 0u; i < 7u; i++)
    {
    	if (bmp_used)
    	{
            result += ((value & (BIT(i))) >> i) * bmp_mgrav_per_bit[i];
    	}
    	else
    	{
            result += ((value & (BIT(i))) >> i) * cma_mgrav_per_bit[i];
    	}
    }

    if (sign == -1)
    	return -result;
    return result;
}


// *************************************************************************************************
// @fn          convert_acceleration_value_to_tenthsOfDegrees
// @brief       Converts acceleration value to tenths of a degree
// @param       s8 value
// @return      s16 decidegrees
// *************************************************************************************************
s16 convert_acceleration_value_to_tenthsOfDegrees(s8 value)
{
	// centigravToTenthsOfDegrees[n] = asin(n / 100) in degrees * 10
	static const s16 centigravToTenthsOfDegrees[] = {
		/* 0 */ 0,
		/* 10 */ 6,
		/* 20 */ 11,
		/* 30 */ 17,
		/* 40 */ 23,
		/* 50 */ 29,
		/* 60 */ 34,
		/* 70 */ 40,
		/* 80 */ 46,
		/* 90 */ 52,
		/* 100 */ 57,
		/* 110 */ 63,
		/* 120 */ 69,
		/* 130 */ 75,
		/* 140 */ 80,
		/* 150 */ 86,
		/* 160 */ 92,
		/* 170 */ 98,
		/* 180 */ 104,
		/* 190 */ 110,
		/* 200 */ 115,
		/* 210 */ 121,
		/* 220 */ 127,
		/* 230 */ 133,
		/* 240 */ 139,
		/* 250 */ 145,
		/* 260 */ 151,
		/* 270 */ 157,
		/* 280 */ 163,
		/* 290 */ 169,
		/* 300 */ 175,
		/* 310 */ 181,
		/* 320 */ 187,
		/* 330 */ 193,
		/* 340 */ 199,
		/* 350 */ 205,
		/* 360 */ 211,
		/* 370 */ 217,
		/* 380 */ 223,
		/* 390 */ 230,
		/* 400 */ 236,
		/* 410 */ 242,
		/* 420 */ 248,
		/* 430 */ 255,
		/* 440 */ 261,
		/* 450 */ 267,
		/* 460 */ 274,
		/* 470 */ 280,
		/* 480 */ 287,
		/* 490 */ 293,
		/* 500 */ 300,
		/* 510 */ 307,
		/* 520 */ 313,
		/* 530 */ 320,
		/* 540 */ 327,
		/* 550 */ 334,
		/* 560 */ 341,
		/* 570 */ 348,
		/* 580 */ 355,
		/* 590 */ 362,
		/* 600 */ 369,
		/* 610 */ 376,
		/* 620 */ 383,
		/* 630 */ 391,
		/* 640 */ 398,
		/* 650 */ 405,
		/* 660 */ 413,
		/* 670 */ 421,
		/* 680 */ 428,
		/* 690 */ 436,
		/* 700 */ 444,
		/* 710 */ 452,
		/* 720 */ 461,
		/* 730 */ 469,
		/* 740 */ 477,
		/* 750 */ 486,
		/* 760 */ 495,
		/* 770 */ 504,
		/* 780 */ 513,
		/* 790 */ 522,
		/* 800 */ 531,
		/* 810 */ 541,
		/* 820 */ 551,
		/* 830 */ 561,
		/* 840 */ 571,
		/* 850 */ 582,
		/* 860 */ 593,
		/* 870 */ 605,
		/* 880 */ 616,
		/* 890 */ 629,
		/* 900 */ 642,
		/* 910 */ 655,
		/* 920 */ 669,
		/* 930 */ 684,
		/* 940 */ 701,
		/* 950 */ 718,
		/* 960 */ 737,
		/* 970 */ 759,
		/* 980 */ 785,
		/* 990 */ 819,
		/* 1000 */ 900,
	};
	s16 centigrav, sign, interpolation, result;
	s16 mgrav = convert_acceleration_value_to_mgrav(value);
	if (mgrav < 0)
	{
		mgrav = -mgrav;
		sign = -1;
	}
	else
		sign = 1;
	centigrav = mgrav / 10;
	interpolation = (centigravToTenthsOfDegrees[centigrav + 1] - centigravToTenthsOfDegrees[centigrav]) * (mgrav % 10) / 10;
    result = centigravToTenthsOfDegrees[centigrav] + interpolation;
    if (sign == -1)
    	result = -result;
    return result;
}

// *************************************************************************************************
// @fn          convert_tenthsOfDegrees_to_grade
// @brief       Converts tenths of a degree to grade
// @param       s16 decidegrees
// @return      s16 Grade (i.e. slope of the line, where a slope of 1 is 45 degrees)
// *************************************************************************************************
s16 convert_tenthsOfDegrees_to_grade(s16 decidegrees)
{
	// degreesToGrade[n] = tan(n) * 100
	static const s16 degreesToGrade[] = {
		/* 0 */ 0,
		/* 1 */ 2,
		/* 2 */ 3,
		/* 3 */ 5,
		/* 4 */ 7,
		/* 5 */ 9,
		/* 6 */ 11,
		/* 7 */ 12,
		/* 8 */ 14,
		/* 9 */ 16,
		/* 10 */ 18,
		/* 11 */ 19,
		/* 12 */ 21,
		/* 13 */ 23,
		/* 14 */ 25,
		/* 15 */ 27,
		/* 16 */ 29,
		/* 17 */ 31,
		/* 18 */ 32,
		/* 19 */ 34,
		/* 20 */ 36,
		/* 21 */ 38,
		/* 22 */ 40,
		/* 23 */ 42,
		/* 24 */ 45,
		/* 25 */ 47,
		/* 26 */ 49,
		/* 27 */ 51,
		/* 28 */ 53,
		/* 29 */ 55,
		/* 30 */ 58,
		/* 31 */ 60,
		/* 32 */ 62,
		/* 33 */ 65,
		/* 34 */ 67,
		/* 35 */ 70,
		/* 36 */ 73,
		/* 37 */ 75,
		/* 38 */ 78,
		/* 39 */ 81,
		/* 40 */ 84,
		/* 41 */ 87,
		/* 42 */ 90,
		/* 43 */ 93,
		/* 44 */ 97,
		/* 45 */ 100,
		/* 46 */ 104,
		/* 47 */ 107,
		/* 48 */ 111,
		/* 49 */ 115,
		/* 50 */ 119,
		/* 51 */ 123,
		/* 52 */ 128,
		/* 53 */ 133,
		/* 54 */ 138,
		/* 55 */ 143,
		/* 56 */ 148,
		/* 57 */ 154,
		/* 58 */ 160,
		/* 59 */ 166,
		/* 60 */ 173,
		/* 61 */ 180,
		/* 62 */ 188,
		/* 63 */ 196,
		/* 64 */ 205,
		/* 65 */ 214,
		/* 66 */ 225,
		/* 67 */ 236,
		/* 68 */ 248,
		/* 69 */ 261,
		/* 70 */ 275,
		/* 71 */ 290,
		/* 72 */ 308,
		/* 73 */ 327,
		/* 74 */ 349,
		/* 75 */ 373,
		/* 76 */ 401,
		/* 77 */ 433,
		/* 78 */ 470,
		/* 79 */ 514,
		/* 80 */ 567,
		/* 81 */ 631,
		/* 82 */ 712,
		/* 83 */ 814,
		/* 84 */ 951,
		/* 85 */ 1143,
		/* 86 */ 1430,
		/* 87 */ 1908,
		/* 88 */ 2864,
		/* 89 */ 5729,
		/* 90 */ 32767,
	};
	s16 degrees, sign, interpolation, result;
	if (decidegrees < 0)
	{
		decidegrees = -decidegrees;
		sign = -1;
	}
	else
		sign = 1;
	degrees = decidegrees / 10;
	interpolation = (degreesToGrade[degrees + 1] - degreesToGrade[degrees]) * (decidegrees % 10) / 10;
    result = degreesToGrade[degrees] + interpolation;
    if (sign == -1)
    	result = -result;
    return result;
}

// *************************************************************************************************
// @fn          is_acceleration_measurement
// @brief       Returns 1 if acceleration is currently measured.
// @param       none
// @return      u8              1 = acceleration measurement ongoing
// *************************************************************************************************
u8 is_acceleration_measurement(void)
{
    return ((sAccel.mode == ACCEL_MODE_ON) && (sAccel.timeout > 0u));
}

// *************************************************************************************************
// @fn          do_acceleration_measurement
// @brief       Get sensor data and store in sAccel struct
// @param       none
// @return      none
// *************************************************************************************************
void do_acceleration_measurement(void)
{
	static s32 lastSample = 0;

    // Get data from sensor
	if (bmp_used)
	{
        bmp_as_get_data((s8 *) sAccel.xyz);
	}
	else
	{
        cma_as_get_data((s8 *) sAccel.xyz);
	}

	moving_sum -= moving_sum_samples[moving_sum_index];
	moving_sum_samples[moving_sum_index] = convert_acceleration_value_to_tenthsOfDegrees(sAccel.xyz[BIKE_AXIS]) - level_ground[BIKE_AXIS];
	// Limit how fast the angle can change. This reduces spurious readings (e.g. going over a bump)
	if (moving_sum_samples[moving_sum_index] > lastSample + 5)
	{
		moving_sum_samples[moving_sum_index] = lastSample + 5;
	}
	else if (moving_sum_samples[moving_sum_index] < lastSample - 5)
	{
		moving_sum_samples[moving_sum_index] = lastSample - 5;
	}
	lastSample = moving_sum_samples[moving_sum_index];
	moving_sum += moving_sum_samples[moving_sum_index];
	if (++moving_sum_index == MOVING_SUM_SAMPLE_COUNT)
		moving_sum_index = 0u;

	if (moving_sum > max_climb)
		max_climb = moving_sum;
	if (moving_sum < max_descent)
		max_descent = moving_sum;

    // Set display update flag
    display.flag.update_acceleration = 1u;
}

// *************************************************************************************************
// @fn          display_acceleration
// @brief       Display routine.
// @param       u8 line                 LINE1
//                              u8 update               DISPLAY_LINE_UPDATE_FULL, DISPLAY_LINE_CLEAR
// @return      none
// *************************************************************************************************
void display_acceleration(u8 line, u8 update)
{
    u8 *str;
    s32 raw_data;
    s16 accel_data;

    // Show warning if acceleration sensor was not initialised properly
    if (!as_ok)
    {
        display_chars(LCD_SEG_L1_2_0, (u8 *) "ERR", SEG_ON);
    }
    else
    {
        // Redraw whole screen
        if (update == DISPLAY_LINE_UPDATE_FULL)
        {
            {
                // Start acceleration sensor
                if (!is_acceleration_measurement())
                {
                    // Start sensor
                	if (bmp_used)
                	{
                        bmp_as_start();
                	}
                	else
                	{
                        cma_as_start();
                	}

                    // Set timeout counter
                    sAccel.timeout = ACCEL_MEASUREMENT_TIMEOUT;

                    // Set mode
                    sAccel.mode = ACCEL_MODE_ON;

                    // Start with Y-axis values
                    sAccel.view_style = BIKE_AXIS;
                }

                // Display decimal point
                //display_symbol(LCD_SEG_L1_DP1, SEG_ON);
            }
        }
        else if (update == DISPLAY_LINE_UPDATE_PARTIAL)
        {
            // Convert X/Y/Z values to mg
            switch (sAccel.view_style)
            {
                case DISPLAY_MAX_CLIMB:
                    raw_data = max_climb;
                    display_symbol(LCD_SYMB_MAX, SEG_ON);
                    break;
                case DISPLAY_MAX_DESCENT:
                    raw_data = max_descent;
                    display_symbol(LCD_SYMB_MAX, SEG_ON);
                    break;
                case DISPLAY_ACCEL:
                default:
                    raw_data = moving_sum;
                    display_symbol(LCD_SYMB_MAX, SEG_OFF);
                    break;
            }
            // Do rounding based on the sign
            if (raw_data > 0)
            	raw_data += MOVING_SUM_SAMPLE_COUNT / 2;
            else
            	raw_data -= MOVING_SUM_SAMPLE_COUNT / 2;
            accel_data = (s16) (raw_data / MOVING_SUM_SAMPLE_COUNT);
            accel_data = convert_tenthsOfDegrees_to_grade(accel_data);

            // Display grade in ddd format
            str = int_to_array(abs(accel_data), 3, 2);
            display_chars(LCD_SEG_L1_2_0, str, SEG_ON);

            // Display sign
            if (accel_data >= 0)
            {
                display_symbol(LCD_SYMB_ARROW_UP, SEG_ON);
                display_symbol(LCD_SYMB_ARROW_DOWN, SEG_OFF);
            }
            else
            {
                display_symbol(LCD_SYMB_ARROW_UP, SEG_OFF);
                display_symbol(LCD_SYMB_ARROW_DOWN, SEG_ON);
            }
            display_symbol(LCD_SYMB_PERCENT, SEG_ON);
        }
        else if (update == DISPLAY_LINE_CLEAR)
        {
            // Stop acceleration sensor
        	if (bmp_used)
        	{
                bmp_as_stop();
        	}
        	else
        	{
                cma_as_stop();
        	}

            // Clear mode
            sAccel.mode = ACCEL_MODE_OFF;

            // Clean up display
            display_symbol(LCD_SEG_L1_DP1, SEG_OFF);
            display_symbol(LCD_SYMB_ARROW_UP, SEG_OFF);
            display_symbol(LCD_SYMB_ARROW_DOWN, SEG_OFF);
            display_symbol(LCD_SYMB_PERCENT, SEG_OFF);
        }
    }
}

